"""
transform information from data block of cif file to class Data
"""
__author__="ikibalin"
__version__="28/08/2019"

from .item import str_to_comment, Item
from .items import Items
from .loop import Loop, find_name_comment_in_block


class Data(object):
    """
    container for objects Items and list of Loop
    """

    def __init__(self, name="", items=None, loops=[], comment=""):
        self.__name = None
        self.__comment = None
        self.__items = None
        self.__loops = None
        self.name = name
        self.comment = comment
        self.items = items
        self.loops = loops

    @property    
    def name(self):
        return self.__name
    @name.setter
    def name(self, x):
        if x is not None:
            name = str(x).strip()
        else:
            name = ""
        self.__name = name
    @property    
    def comment(self):
        return self.__comment
    @comment.setter
    def comment(self, x):
        comment = str_to_comment(x)
        self.__comment = comment
    @property    
    def comments(self):
        return self.items.comment

    @property    
    def items(self):
        return self.__items
    @items.setter
    def items(self, x):
        if isinstance(x, Items):
            res = x
        elif isinstance(x, str):
            cif_values = Items()
            flag = cif_values.take_from_string(x)
            if flag:
                res = cif_values
            else:
                self._show_message("Can not convert string to object 'Items'")
                res = None
        elif x is None:
            res = None
        else:
            try:
                flag_1 = isinstance(x[0], Item)
                flag_2 = isinstance(x[0], str)
            except:
                flag_1, flag_2 = False, False
            cif_values = Items()
            if flag_1:
                cif_values.items = x
                res = cif_values
            elif flag_2:
                try:
                    cif_values.items = "\n".join(x)
                    res = cif_values
                except:
                    self._show_message("Can not convert list of strings to object 'Items'")
                    res = None
            else:
                self._show_message("Can not define the type of input data to convert it to object 'Items'")
                res = None
        self.__items = res

    @property    
    def loops(self):
        return tuple(self.__loops)
    @loops.setter
    def loops(self, l_x):
        if isinstance(l_x, Loop):
            l_x_in = [l_x]
        elif isinstance(l_x, str):
            l_x_in = [l_x]
        l_cif_loop = []
        for x in l_x:
            if isinstance(x, Loop):
                l_cif_loop.append(x)
            elif isinstance(x, str):
                cif_loop = Loop()
                flag_out = cif_loop.take_from_string(x)
                if flag_out:
                    l_cif_loop.append(cif_loop)
                else:
                    self._show_message("Can not convert list of strings to object 'Loop'")
            else:
                self._show_message("Can not define the type of input data to convert it to object 'Loop'")
        self.__loops = l_cif_loop

    @property
    def is_values(self):
        return (self.items is not None)

    
    def is_value(self, key_):
        str_1 = delete_name_from_prefix(key_, self.name)

        flag_1 = False
        if self.is_values:
            flag_1 = self.items.is_value(str_1)
        if flag_1:
            return flag_1
        for cif_loop in self.loops:
            flag_1 = cif_loop.is_value(str_1)
            if flag_1:
                return flag_1
        return flag_1

    def is_prefix(self, key_):
        str_1 = delete_name_from_prefix(key_, self.name)

        flag_1 = False
        if self.is_values:
            flag_1 = self.items.is_prefix(str_1)
        if flag_1:
            return flag_1
        for cif_loop in self.loops:
            flag_1 = cif_loop.is_prefix(str_1)
            if flag_1:
                return flag_1
        return flag_1

    def add_item(self, x):
        if self.items is None:
            self.items = Items(items=[x])
        else:
            self.items.add_item(x)

    def add_loop(self, x):
        if isinstance(x, Loop):
            x_in = [x]
        elif isinstance(x, str):
            _1 = Loop()
            flag = _1.take_from_string(x)
            if flag:
                x_in = [_1]
        else:
            x_in = []
            self._show_message("Introduced object was not identified as a loop")
        self.__loops.extend(x_in)

    def __repr__(self):
        res = str(self)
        return res

    def __str__(self):
        ls_out = ["data_{:} {:}\n".format(self.name, self.comment)]
        ls_out.append("\n")
        if self.is_values:
            ls_out.append(str(self.items))
        for cif_loop in self.loops:
            ls_out.append("\n\n")
            ls_out.append(str(cif_loop))
        return "".join(ls_out)

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    def __getitem__(self, key_):
        str_1 = delete_name_from_prefix(key_, self.name)

        flag_1 = self.items.is_value(str_1)
        flag_2 = self.items.is_prefix(str_1)
        res = None
        if flag_1:
            res = self.items[str_1]
        elif flag_2:
            res = self.items.items_with_prefix(str_1)
        else:
            for cif_loop in self.loops:
                flag_1 = cif_loop.is_value(str_1)
                flag_2 = (cif_loop.prefix == str_1)
                if flag_1:
                    res = cif_loop[str_1]
                    break
                elif flag_2:
                    res = cif_loop
                    break
        if res is None:
            self._show_message("Item is not found")
        return res

    def take_from_string(self, string):
        if isinstance(string, str):
            l_string = string.split("\n")
        else:
            l_string = string

        name, comment, i_data = find_name_comment_in_block(l_string, "data_")
        self.name = name
        self.comment = comment

        flag_loop_read_name = False
        flag_loop_read_array = False
        l_string_values, l_string_loops = [], []
        l_string_loop = []
        for line in l_string[(i_data+1):]:
            str_1 = line.strip()
            cond_1 = any([str_1.lower().startswith(key_word) for key_word in ["data_", "global_"]])
            cond_2 = (str_1 == "")
            cond_3 = str_1.lower().startswith("loop_")
            cond_4 = str_1.lower().startswith("_")
            cond_5 = str_1.lower().startswith("#")
            if cond_1:
                break
            elif cond_2:
                pass
            elif cond_3:
                if l_string_loop != []:
                    l_string_loops.append("\n".join(l_string_loop))
                    l_string_loop = []
                flag_loop_read_name = True
                flag_loop_read_array = False
            elif (cond_4 | cond_5):
                if flag_loop_read_name:
                    l_string_loop.append(str_1)
                else:
                    flag_loop_read_name = False
                    flag_loop_read_array = False
                    l_string_values.append(str_1)
            else:
                if (flag_loop_read_name | flag_loop_read_array):
                    flag_loop_read_name = False
                    flag_loop_read_array = True
                    l_string_loop.append(str_1)
                else:
                    l_string_values.append(str_1)
        self.items = "\n".join(l_string_values)
        if l_string_loop != []:
            l_string_loops.append("\n".join(l_string_loop))
        if l_string_loops != []:
            self.loops = l_string_loops
        return True

    def __add__(self, x):
        if isinstance(x, Data):
            if self.name != x.name:
                self._show_message("Data blocks should have the same name")
                return None
            res = Data()
            res.name = self.name
            res.comment = self.comment

            res.items = self.items
            if x.items is not None:
                if res.items is None:
                    res.items = x.items 
                else: 
                    res.items = res.items + x.items
            l_loops = [_ for _ in self.loops]
            l_loops.extend([_ for _ in x.loops])
            res.loops =  l_loops
        else:
            res = None
            self._show_message("Sum operation is defined only for Data type")
        return res
    def __radd__(self, x):
        return self+x



def delete_name_from_prefix(key_: str, s_name: str):
    str_1 = key_.strip().lower()
    if str_1.startswith(s_name.lower()):
        str_1 = str_1[len(s_name):]
    elif str_1.startswith("_"+s_name.lower()):
        str_1 = str_1[(1+len(s_name)):]
    if not(str_1.startswith("_")):
        str_1 = "_"+str_1
    return str_1





"""
container of Item
"""
__author__="ikibalin"
__version__="2019/09/12"

from .item import Item, str_to_comment

class Items(object):
    """
    container for objects Item
    """

    def __init__(self, items=[], comment=""):
        self.__items = None
        self.__comment = None
        self.comment = comment
        self.items = items

    @property    
    def items(self):
        return tuple(self.__items)
    @items.setter
    def items(self, l_x):
        l_res = []
        try:
            if isinstance(l_x, str):
                l_x_in = [l_x]
            else:
                _ = [x for x in l_x]
                l_x_in = l_x
        except:
            l_x_in = [l_x]
        for x in l_x_in:
            if isinstance(x, Item):
                l_res.append(x)
            else:
                item = Item()
                flag_out = item.take_from_string(str(x))
                if flag_out:
                    l_res.append(item)
        self.__items = l_res
    @property    
    def comment(self):
        return self.__comment
    @comment.setter
    def comment(self, x):
        if isinstance(x, str):
            x_in = str_to_comment(x)
        else:
            x_in = str("\n".join([str_to_comment(hh) for hh in x]))
        self.__comment = x_in

    @property    
    def names(self):
        return tuple([hh.name for hh in self.items])
    def __repr__(self):
        res = str(self)
        return res

    def __str__(self):
        if self.is_defined:
            ls_out = [self.comment]
            ls_out.extend([str(item) for item in self.items])
            return "\n".join(ls_out)
        return "The object 'Items' is not defined"

    @property
    def is_defined(self):
        cond = (self.items is not None)
        return cond

    def is_value(self, key_):
        str_1 = key_.strip().lower()
        if not(str_1.startswith("_")):
            str_1 = "_"+str_1
        l_name = [hh.name for hh in self.items]
        cond = (str_1 in l_name)
        return cond

    def is_prefix(self, key_):
        str_1 = key_.strip().lower()
        if not(str_1.startswith("_")):
            str_1 = "_"+str_1
        l_res = self.items_with_prefix(str_1)
        cond = (len(l_res) > 0)
        return cond

    def add_item(self, x):
        if isinstance(x, Item):
            x_in = [x]
        elif isinstance(x, str):
            _1 = Item()
            flag = _1.take_from_string(x)
            if flag:
                x_in = [_1]
        else:
            x_in = []
            self._show_message("Introduced object was not identified as an item")
        self.__items.extend(x_in)

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return iter(self.items)

    def __getitem__(self, key_):
        res = None
        str_1 = key_.lower().strip()
        if not(str_1.startswith("_")):
            str_1 = "_"+str_1
        for item in self.items:
            if item.name == str_1:
                res = item
                break
        l_res = self.items_with_prefix(str_1)
        if ((res is None) & (len(l_res) > 0)):
            res = l_res
        elif (res is None):
            self._show_message("Name '{:}' is not found".format(key_))
        return res

    def items_with_prefix(self, s_type: str):
        str_1 = s_type.lower().strip()
        if not(str_1.startswith("_")):
            str_1 = "_"+str_1
        l_res = Items(items=[])
        l_type = str_1.split("_")
        len_l_type = len(l_type)
        for item in self.items:
            l_type_name = item.name.split("_")
            if len_l_type <= len(l_type_name):
                flag = all([hh_1 == hh_2 for hh_1, hh_2 in zip(l_type, l_type_name[:len_l_type])])
                if flag:
                    l_res.add_item(item)
        return l_res

    def take_from_string(self, string):
        if isinstance(string, str):
            l_string_in = string.split("\n")
        else:
            l_string_in = string
        l_value_string, l_comment = [], []
        s_val = []
        flag_in = False
        for i_line, line in enumerate(l_string_in):
            if line.strip().startswith(";"):
                s_val.append(line.strip())
                flag_in = not(flag_in)
            elif line.strip().startswith("#"):
                if flag_in:
                    s_val.append(line.strip())
                else:
                    l_comment.append(line.strip())
            elif line.strip().startswith("_"):
                if flag_in:
                    s_val.append(line.strip())
                else:
                    if s_val != []:
                        l_value_string.append("\n".join(s_val))
                    s_val = [line.strip()]
            elif flag_in:
                s_val.append(line.strip())
        if s_val != []:
            l_value_string.append("\n".join(s_val))
        self.comment = "\n".join(l_comment)
        self.items = l_value_string
        return True

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    def __add__(self, x):
        if isinstance(x, Items):
            l_items = [_ for _ in self.items]
            l_items.extend([_ for _ in x.items])
            res = Items(items = l_items)
            if self.comment != "":
                res.comment = self.__comment
            if x.comment != "":
                if res.comment != "":
                    res.comment += "\n"+x.comment
                else:
                    res.comment = x.comment
        else:
            res = None
            self._show_message("Sum operation is defined only for Items type")
        return res
    def __radd__(self, x):
        return self+x



name = "pycifstar"

from .item import Item
from .items import Items
from .loop import Loop
from .data import Data
from .global_ import Global

def read_star_file(f_name):
    return to_global(f_name)

def read_file(f_name):
    return to_global(f_name)

def to_global(f_name):
    fid = open(f_name, "r")
    string = fid.read()
    fid.close()
    global_ = Global()
    flag = global_.take_from_string(string)
    if not flag:
        print("Error at file reading")
    return global_

def to_data(f_name):
    fid = open(f_name, "r")
    string = fid.read()
    fid.close()
    data = Data()
    flag = data.take_from_string(string)
    if not flag:
        print("Error at file reading")
    return data

def to_loop(f_name):
    fid = open(f_name, "r")
    string = fid.read()
    fid.close()
    loop = Loop()
    flag = loop.take_from_string(string)
    if not flag:
        print("Error at file reading")
    return loop

def to_values(f_name):
    fid = open(f_name, "r")
    string = fid.read()
    fid.close()
    items = Items()
    flag = items.take_from_string(string)
    if not flag:
        print("Error at file reading")
    return items



"""
transform information from loop of cif file to class Loop
"""
__author__="ikibalin"
__version__="28/08/2019"

from .item import print_string, str_to_comment

class Loop(object):
    """
    class to store loop from .cif
    """

    def __init__(self, name=None, names = None, values = None, comment=None, names_comment = None, values_comment = None):
        self.__name = None
        self.__comment = None
        self.__names = None
        self.__values = None
        self.__values_comment = None
        self.__names_comment = None
        self.name = name
        self.names = names
        self.values = values
        self.comment = comment
        self.names_comment = names_comment
        self.values_comment = values_comment

    @property    
    def name(self):
        return self.__name
    @name.setter
    def name(self, x):
        if x is not None:
            name = str(x).strip()
        else:
            name = ""
        self.__name = name
    @property    
    def comment(self):
        return self.__comment
    @comment.setter
    def comment(self, x):
        comment = str_to_comment(x)
        self.__comment = comment
    @property    
    def values(self):
        return self.__cif_value_array
    @values.setter
    def values(self, ll_x):
        if ll_x is not None:
            try:
                try:
                    self.__cif_value_array = [[str(x).strip().strip("\"").strip("'") for x in l_x] for l_x in ll_x]
                except:
                    self.__cif_value_array = [[str(l_x).strip().strip("\"").strip("'")] for l_x in ll_x]
            except:
                self.__cif_value_array = [[str(ll_x).strip()]]
        else:
            self.__cif_value_array = None
    @property    
    def table(self):
        return self.values
    @table.setter
    def table(self, ll_x):
        self.values= ll_x


    @property    
    def names(self):
        return self.__names if self.__names is None else tuple(self.__names)
    @names.setter
    def names(self, l_x):
        if l_x is not None:
            try:
                l_name = [str(x).strip().lower() for x in l_x]
            except:
                l_name = [str(l_x).strip().lower()]
        else:
            l_name = None
        self.__names = l_name
    @property    
    def values_comment(self):
        return self.__values_comment
    @values_comment.setter
    def values_comment(self, l_x):
        if l_x is not None:
            
            try:
                self.__values_comment = [str_to_comment(x) for x in l_x]
            except:
                self.__values_comment = [str_to_comment(l_x)]
        else:
            if self.values is not None:
                nx, ny = self.size_values
                if ny > 0:
                    self.__values_comment = ny*[""]
            self.__values_comment = None

    @property    
    def table_comment(self):
        return self.values_comment
    @table_comment.setter
    def table_comment(self, ll_x):
        self.values_comment= ll_x

    @property    
    def names_comment(self):
        return self.__names_comment
    @names_comment.setter
    def names_comment(self, l_x):
        if l_x is not None:
            try:
                l_res = [str_to_comment(x) for x in l_x]
            except:
                l_res = [str_to_comment(l_x)]
            if len(l_res) == self.len_names:
                self.__names_comment = l_res
            else:
                self._show_message("Lens of name comment is not in agreement with lens of names")
        else:
            if self.len_names > 0:
                self.__names_comment = self.len_names*['']
            else:
                self.__names_comment = None

    @property    
    def len_names(self):
        if self.names is None:
            res = 0
        else:
            res = len(self.__names)
        return res
    @property    
    def size_values(self):
        if self.values is None:
            nx, ny = 0, 0
        else:
            ll_x = self.values
            ny = len(ll_x)
            nx = len(ll_x[0])
            if ny > 1:
                for l_x in ll_x[1:]:
                    if len(l_x) != nx:
                        self._show_message("Size of data values is not constant")
                        nx = None
                        break
        return (nx, ny)

    @property    
    def size_table(self):
        return self.size_values

    @property    
    def _check_object(self):
        if self.is_defined:
            nx_name = self.len_names
            nx, ny = self.size_values
            if nx is None:
                return False
            if nx != nx_name:
                self._show_message("Disagreement between number of name and number of elements in table")
                return False
            l_values_comment = self.values_comment
            if  l_values_comment is None:
                l_values_comment = ny*['']
            if len(l_values_comment) < ny:
                l_values_comment.extend((ny-len(l_values_comment))*[""])
            self.values_comment = l_values_comment
            return True
        return False
    
    def __str__(self):
        flag = self._check_object
        if not(flag):
            self._show_message("Data introduced incorrectly")
            return ""
        l_line_1 = ["loop_"]
        if self.name is not None:
            l_line_1.append("{:}".format(self.name))
        if self.comment is not None:
            l_line_1.append(" {:}".format(self.comment))
        if self.is_defined:
            ls_out = ["".join(l_line_1)]
            if self.len_names > 0:
                ls_out.extend(["{:}  {:}".format(hh_1, hh_2) for hh_1, hh_2 in zip(self.names, self.names_comment)])
                ll_x = self.values
                l_values_comment = self.values_comment
                for l_x, comment in zip(ll_x, l_values_comment):
                    l_line_1 = ["{:}".format(print_string(x)) for x in l_x]
                    l_line_1.append(comment)
                    ls_out.append(" ".join(l_line_1))
        return "\n".join(ls_out)

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    def __repr__(self):
        res = str(self)
        return res

    @property
    def is_defined(self):
        cond = ((self.names is not None) & (self.values is not None))
        return cond
    @property
    def is_value_comment(self):
        cond = (self.values_comment is not None)
        return cond
    @property
    def prefix(self):
        if self.is_defined:
            l_name = self.names
            l_first_type = l_name[0].split("_")
            if self.len_names == 1:
                res = "_".join(l_first_type)
            else: #if more than 1
                ind_smallest = len(l_name[0].split("_"))
                for name in l_name[1:]:
                    l_type = name.split("_")
                    len_1 = min([len(l_first_type), len(l_type)])
                    l_help = [hh_1 == hh_2 for hh_1, hh_2 in zip(l_first_type[:len_1], l_type[:len_1])]
                    if False in l_help:
                        ind_0 = l_help.index(False)
                    else:
                        ind_0 = len(l_help)
                    ind_smallest = min([ind_0, ind_smallest])
                res = "_".join(l_first_type[:ind_smallest])
        else:
            self._show_message("Object is not defined")
            res = None
        return res
    def is_prefix(self, key_):
        str_1 = key_.strip().lower()
        prefix = self.prefix
        cond = (str_1==prefix)
        return cond

    def __getitem__(self, key_):
        if self.is_defined:
            str_1 = key_.strip().lower()
            if not(str_1.startswith("_")):
                str_1 = "_"+str_1
            if str_1 in self.names:
                key_ind = self.names.index(str_1)
                ll_x = self.values
                l_res = [l_x[key_ind] for l_x in ll_x]
                return l_res
            else:
                self._show_message("The name {:} is not found".format(key_))    
        else:
            self._show_message("Object is not defined")

    def is_value(self, key_):
        str_1 = key_.strip().lower()
        if not(str_1.startswith("_")):
            str_1 = "_"+str_1
        cond = (str_1 in self.names)
        return cond

    def take_from_string(self, string):
        """
        should be formatted as

        loop_name # comment of loop
        _coeff_1 # comment name
        _coeff_2
        _coeff_3
        1  2  3  # comment 1
        5  4  1
        7  8  6  # other comment
        """
        if isinstance(string, str):
            l_string = string.split("\n")
        else:
            l_string = string
        
        name, comment, i_loop = find_name_comment_in_block(l_string, "loop_")
        self.name = name
        self.comment = comment

        l_name, l_name_comment = [], []
        for i_line, line in enumerate(l_string[i_loop+1:]):
            str_1 = line.strip()
            if str_1.startswith("_"):
                ind_1= str_1.find(" ")
                if ((ind_1 != -1)):
                    l_name.append(str_1[:ind_1])
                    l_name_comment.append(str_1[ind_1:])
                else:
                    l_name.append(str_1)
                    l_name_comment.append("")
            else:
                i_table = i_line+i_loop+1
                break
        len_name = len(l_name)

        l_values, l_values_comment = [], []
        for i_line, line in enumerate(l_string[i_table:]):
            str_1 = line.strip()
            cond = any([str_1.startswith(key_word) for key_word in ["loop_", "data_", "global_", "_", "#"]])
            if (cond | (str_1 == '')):
                break
            ind_1 = str_1.find("#")
            if ind_1 == -1:
                comment = ""
                line_value = str_1
            else:
                comment = str_1[ind_1:]
                line_value = str_1[:ind_1]

            l_value = smart_split(line_value)
            if len(l_value) < len_name:
                l_value.extend((len_name-len(l_value))*["."])

            l_values.append(l_value[:len_name])
            l_values_comment.append(comment)
                
        self.names = l_name
        self.names_comment = l_name_comment
        self.values = l_values
        self.values_comment = l_values_comment
        return True



def smart_split(line):
    """
    split string like:
    "C C 0.0033 0.0016 'International Tables Vol C Tables 4.2.6.8 and 6.1.1.4'"
    in the list like:
    ['C', 'C', '0.0033', '0.0016', 'International Tables Vol C Tables 4.2.6.8 and 6.1.1.4']
    """
    flag_in = False
    l_val, val = [], []
    for hh in line.strip():
        if (hh == " ") & (not flag_in):
            if val != []:
                l_val.append("".join(val))
            val = []
        elif (hh == " ") & (flag_in):
            val.append(hh)
        elif ((hh == "'")|(hh == "\"")):
            flag_in = not flag_in
        else:
            val.append(hh)
    if val != []:
        l_val.append("".join(val))
    return l_val


def find_name_comment_in_block(l_string, block_name):
    block_name_l = block_name.strip().lower()
    if not(block_name_l.endswith("_")):
        block_name_l = block_name_l + "_"
    len_bn = len(block_name_l)
    l_help = [i_line for i_line, line in enumerate(l_string) if line.strip().lower().startswith(block_name_l)]
    if len(l_help) > 0:
        i_block = l_help[0]
        str_1 = l_string[i_block].strip()
        ind_1= str_1.find(" ")
        if ((ind_1 != -1)):
            name = str_1[len_bn:ind_1]
            comment = str_1[ind_1:]
        else:
            name = str_1[len_bn:]
            comment = ""
    else:
        i_block = -1
        name = ""
        comment = ""
    return name, comment, i_block



"""
transform information from global block of cif file to class Global
"""
__author__="ikibalin"
__version__="29/08/2019"

from .item import str_to_comment, Item
from .items import Items
from .loop import Loop, find_name_comment_in_block
from .data import Data, delete_name_from_prefix


class Global(object):
    """
    container for objects Items, list of Loop and Data
    """

    def __init__(self, name="", items=None, loops=[], datas=[], comment=""):
        self.__name = None
        self.__comment = None
        self.__items = None
        self.__loops = None
        self.__datas = None
        self.name = name
        self.comment = comment
        self.items = items
        self.loops = loops
        self.datas = datas

    @property    
    def name(self):
        return self.__name
    @name.setter
    def name(self, x):
        if x is not None:
            name = str(x).strip()
        else:
            name = ""
        self.__name = name
    @property    
    def comment(self):
        return self.__comment
    @comment.setter
    def comment(self, x):
        comment = str_to_comment(x)
        self.__comment = comment

    @property    
    def comments(self):
        return self.items.comment

    @property    
    def items(self):
        return self.__items
    @items.setter
    def items(self, x):
        if isinstance(x, Items):
            res = x
        elif isinstance(x, str):
            cif_values = Items()
            flag = cif_values.take_from_string(x)
            if flag:
                res = cif_values
            else:
                self._show_message("Can not convert string to object 'Items'")
                res = None
        elif x is None:
            res = None
        else:
            try:
                flag_1 = isinstance(x[0], Item)
                flag_2 = isinstance(x[0], str)
            except:
                flag_1, flag_2 = False, False
            cif_values = Items()
            if flag_1:
                cif_values.items = x
                res = cif_values
            elif flag_2:
                try:
                    cif_values.items = "\n".join(x)
                    res = cif_values
                except:
                    self._show_message("Can not convert list of strings to object 'Items'")
                    res = None
            else:
                self._show_message("Can not define the type of input data to convert it to object 'Items'")
                res = None
        self.__items = res

    @property    
    def loops(self):
        return tuple(self.__loops)
    @loops.setter
    def loops(self, l_x):
        if isinstance(l_x, Loop):
            l_x_in = [l_x]
        elif isinstance(l_x, str):
            l_x_in = [l_x]
        l_cif_loop = []
        for x in l_x:
            if isinstance(x, Loop):
                l_cif_loop.append(x)
            elif isinstance(x, str):
                cif_loop = Loop()
                flag_out = cif_loop.take_from_string(x)
                if flag_out:
                    l_cif_loop.append(cif_loop)
                else:
                    self._show_message("Can not convert list of strings to object 'Loop'")
            else:
                self._show_message("Can not define the type of input data to convert it to object 'Loop'")
        self.__loops = l_cif_loop


    @property    
    def datas(self):
        return self.__datas
    @datas.setter
    def datas(self, l_x):
        if isinstance(l_x, Data):
            l_x_in = [l_x]
        elif isinstance(l_x, str):
            l_x_in = [l_x]
        l_cif_data = []
        for x in l_x:
            if isinstance(x, Data):
                l_cif_data.append(x)
            elif isinstance(x, str):
                data = Data()
                flag_out = data.take_from_string(x)
                if flag_out:
                    l_cif_data.append(data)
                else:
                    self._show_message("Can not convert list of strings to object 'Data'")
            else:
                self._show_message("Can not define the type of input data to convert it to object 'Data'")
        self.__datas = l_cif_data


    @property
    def is_values(self):
        return (self.items is not None)

    def __repr__(self):
        res = str(self)
        return res

    def __str__(self):
        ls_out = ["global_{:} {:}\n".format(self.name, self.comment)]
        ls_out.append("\n")
        if self.is_values:
            ls_out.append(str(self.items))
        for cif_loop in self.loops:
            ls_out.append("\n\n")
            ls_out.append(str(cif_loop))
        for data in self.datas:
            ls_out.append("\n\n")
            ls_out.append(str(data))
        return "".join(ls_out)

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    def __getitem__(self, key_):
        str_1 = delete_name_from_prefix(key_, self.name)

        flag_1 = self.items.is_value(str_1)
        flag_2 = self.items.is_prefix(str_1)
        res = None
        if flag_1:
            res = self.items[str_1]
        elif flag_2:
            res = self.items.items_with_prefix(str_1)
        if res is None:
            for cif_loop in self.loops:
                flag_1 = cif_loop.is_value(str_1)
                flag_2 = cif_loop.is_prefix(str_1)
                if flag_1:
                    res = cif_loop[str_1]
                    break
                elif flag_2:
                    res = cif_loop
                    break
        if res is None:
            for data in self.datas:
                flag_1 = data.is_value(str_1)
                flag_2 = data.is_prefix(str_1)
                flag_3 = (str_1 == ("_"+data.name.lower()))
                if (flag_1 | flag_2):
                    res = data[str_1]
                    break
                if flag_3:
                    res = data
                    break
        if res is None:
            self._show_message("Item is not found")
        return res

    def is_value(self, key_):
        str_1 = delete_name_from_prefix(key_, self.name)

        flag_1 = False
        if self.is_values:
            flag_1 = self.items.is_value(str_1)
        if flag_1:
            return flag_1
        for cif_loop in self.loops:
            flag_1 = cif_loop.is_value(str_1)
            if flag_1:
                return flag_1
        for data in self.datas:
            flag_1 = data.is_value(str_1)
            if flag_1:
                return flag_1
        return flag_1

    def is_prefix(self, key_):
        str_1 = delete_name_from_prefix(key_, self.name)

        flag_1 = False
        if self.is_values:
            flag_1 = self.items.is_prefix(str_1)
        if flag_1:
            return flag_1
        for cif_loop in self.loops:
            flag_1 = cif_loop.is_prefix(str_1)
            if flag_1:
                return flag_1
        for data in self.datas:
            flag_1 = data.is_prefix(str_1)
            if flag_1:
                return flag_1
        return flag_1



    def take_from_string(self, string):
        if isinstance(string, str):
            l_string = string.split("\n")
        else:
            l_string = string

        name, comment, i_global = find_name_comment_in_block(l_string, "global_")
        self.name = name
        self.comment = comment

        flag_loop_read_name = False
        flag_loop_read_array = False
        flag_data_read = False
        l_string_values, l_string_loops = [], []
        l_string_loop = []
        i_data = -1
        for i_line, line in enumerate(l_string[(i_global+1):]):
            str_1 = line.strip()
            cond_1 = any([str_1.lower().startswith(key_word) for key_word in ["data_", "global_"]])
            cond_2 = (str_1 == "")
            cond_3 = str_1.lower().startswith("loop_")
            cond_4 = str_1.lower().startswith("_")
            cond_5 = str_1.lower().startswith("#")
            if cond_1:
                if str_1.lower().startswith("data_"):
                    i_data = i_line+i_global+1
                break
            elif cond_2:
                pass
            elif cond_3:
                if l_string_loop != []:
                    l_string_loops.append("\n".join(l_string_loop))
                    l_string_loop = []
                flag_loop_read_name = True
                flag_loop_read_array = False
            elif (cond_4 | cond_5):
                if flag_loop_read_name:
                    l_string_loop.append(str_1)
                else:
                    flag_loop_read_name = False
                    flag_loop_read_array = False
                    l_string_values.append(str_1)
            else:
                if (flag_loop_read_name | flag_loop_read_array):
                    flag_loop_read_name = False
                    flag_loop_read_array = True
                    l_string_loop.append(str_1)
                else:
                    l_string_values.append(str_1)
        
        self.items = "\n".join(l_string_values)
        if l_string_loop != []:
            l_string_loops.append("\n".join(l_string_loop))
        if l_string_loops != []:
            self.loops = l_string_loops

        if (i_data != -1):
            l_string_datas, l_string_data = [], []
            for line in l_string[i_data:]:
                str_1 = line.strip()
                cond_1 = str_1.lower().startswith("global_")
                cond_2 = str_1.lower().startswith("data_") 
                if cond_1:
                    break
                elif cond_2:
                    if l_string_data != []:
                        l_string_datas.append("\n".join(l_string_data))
                    l_string_data = [str_1]
                else:
                    l_string_data.append(str_1)

            if l_string_data != []:
                l_string_datas.append("\n".join(l_string_data))
            if l_string_datas != []:
                self.datas = l_string_datas

        return True

    def __add__(self, x):
        if isinstance(x, Global):
            res = Global()
            if x.name != "":
                res.name = x.name
                res.comment = x.comment
            else:
                res.name = self.name
                res.comment = self.comment
            res.items = self.items
            if x.items is not None:
                if res.items is None:
                    res.items = x.items 
                else: 
                    res.items = res.items + x.items
            l_loops = [_ for _ in self.loops]
            l_loops.extend([_ for _ in x.loops])
            res.loops =  l_loops

            l_name_1 = [_1.name for _1 in self.datas]
            l_name_2 = [_1.name for _1 in x.datas]
            l_name_12 = list(set(l_name_1+l_name_2))
            l_res_data = []
            for _1 in l_name_12:
                if (_1 in l_name_1) & (_1 in l_name_2):
                    ind_1 = l_name_1.index(_1)
                    ind_2 = l_name_2.index(_1)
                    res_data = self.datas[ind_1]+x.datas[ind_2]
                elif (_1 in l_name_1):
                    ind_1 = l_name_1.index(_1)
                    res_data = self.datas[ind_1]
                else:
                    ind_2 = l_name_2.index(_1)
                    res_data = x.datas[ind_2]
                l_res_data.append(res_data)
            res.datas = l_res_data
        else:
            res = None
            self._show_message("Sum operation is defined only for Global type")
        return res
    def __radd__(self, x):
        return self+x

    def to_file(self, f_name):
        with open(f_name, "w") as fid:
            fid.write(str(self))


#!/usr/bin/env python

# Copyright (c) 2018-2019 Iurii Kibalin   
# https://github.com/ikibalin/pycifstar  
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# 
# * Neither the name "pycifstar" nor the names of its contributors may
#   be used to endorse or promote products derived from this software
#   without specific prior written permission.
# 
# This software is provided by the copyright holders and contributors "as
# is" and any express or implied warranties, including, but not limited
# to, the implied warranties of merchantability and fitness for a
# particular purpose are disclaimed. In no event shall the copyright owner
# or contributors be liable for any direct, indirect, incidental, special,
# exemplary, or consequential damages (including, but not limited to,
# procurement of substitute goods or services; loss of use, data, or
# profits; or business interruption) however caused and on any theory of
# liability, whether in contract, strict liability, or tort (including
# negligence or otherwise) arising in any way out of the use of this
# software, even if advised of the possibility of such damage.


from setuptools import setup, find_packages


long_description = '''\
PyCifStar is  a class library for  data manipulation  provided in 
the Self Defining Text Archival and Retrieval (STAR). The STAR syntax 
provides a way for simple, easy-to-comprehend, flexible and extensible 
data exchange. 

It is written in Python.

For more details, please go to the `github`_.

.. _`github`: https://github.com/ikibalin/pycifstar'''

#with open("readme.rst", 'r') as f:
#    long_description = f.read()

setup(
    name='pycifstar',
    version='0.2.1',
    description='PyCifStar is  a class library for  data manipulation  provided in the CIF/STAR File.',
    long_description = long_description,
    author='Iurii Kibalin',
    author_email='yurikibalin@outlook.com',
    url = 'https://github.com/ikibalin/pycifstar',
    license          = 'MIT License',
    keywords         = ['STAR', 'CIF'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],    
    packages=find_packages(),  #same as name
    install_requires=[]
)


"""
transform information from one variable of .cif file to Item
"""
__author__="ikibalin"
__version__="2019/09/12"



class Item(object):
    """
    class to store one line from .star, .cif
    """

    def __init__(self, name=None, value=None, comment=""):
        self.__name = None
        self.__value = None
        self.__comment = None
        self.name = name
        self.value = value
        self.comment = comment

    @property    
    def name(self):
        return self.__name
    @name.setter
    def name(self, x):
        if x is not None:
            cif_name = str(x).strip().lower()
            if not(cif_name.startswith("_")):
                cif_name = "_"+cif_name
        else:
            cif_name = None
        self.__name = cif_name

    @property    
    def tag(self):
        return self.name
    @tag.setter
    def tag(self, x):
        self.name = x


    @property    
    def value(self):
        return self.__value
    @value.setter
    def value(self, x):
        if x is not None:
            self.__value = str(x).strip().strip("\"").strip("'")
        else:
            self.__value = None
    @property    
    def comment(self):
        return self.__comment
    @comment.setter
    def comment(self, x):
        comment = str_to_comment(x)
        self.__comment = comment
    def __float__(self):
        return float(self.value)


    def __repr__(self):
        res = str(self)
        return res

    def __str__(self):
        if self.is_defined:
            if self.is_comment:
                comment = self.comment
            else:
                comment= ""
            if self.is_value_lines:
                ls_out = ["{:} {:}".format(self.name, comment)]
                ls_out.append("\n;\n")
                ls_out.append(self.value)
                ls_out.append("\n;")
            else:
                ls_out = ["{:} {:}".format(self.name, print_string(self.value))]
                ls_out.append(" {:}".format(comment))
            return "".join(ls_out)
        return "Item is not defined"

    @property
    def is_defined(self):
        cond = ((self.name is not None) & (self.value is not None))
        return cond
    @property
    def is_value_lines(self):
        cond = len(self.value.split("\n")) > 1
        return cond
    @property
    def is_comment(self):
        cond = (self.comment is not None)
        return cond


    def take_from_string(self, string):
        is_strings = len(string.strip().split("\n")) > 1
        if is_strings:
            first_string = string.strip().split("\n")[0]
        else:
            first_string = string

        ind_1 = first_string.find("#")
        if ind_1 == -1:
            self.comment = ""
            str_1 = first_string.strip()
        else:
            self.comment = first_string[ind_1:]
            str_1 = first_string[:ind_1].strip()
        
        l_help = str_1.split(" ")

        if len(l_help) == 0:
            self._show_message("Undefined format for string\n")
            return False

        self.name = l_help[0]
        if not(is_strings):
            if len(l_help)==1:
                self.value = "."
            else:
                self.value = str_1[len(l_help[0]):].strip().strip("\"").strip("'")
            return True
        else:
            l_string = string.strip().split("\n")
            str_1 = l_string[1].strip()
            if str_1.startswith(";"):
                l_help = [i_line+2 for i_line, line in enumerate(l_string[2:]) if line.strip().startswith(";")]
                if len(l_help) > 0:
                    value = str_1.lstrip(";")+"\n"+"\n".join(l_string[2:l_help[0]])
                else:
                    value = str_1.lstrip(";")+"\n"+"\n".join(l_string[2:])
                self.value = value
                return True
            else:
                self._show_message("Can not find ';' for multistring variable")
                return False
        return True

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)


def print_string(str_1):
    str_2 = str_1.strip("\"").strip("'")
    if len(str_2.split()) > 1:
        res = "'{:}'".format(str_2)
    else:
        res = "{:}".format(str_2)
    return res

def str_to_comment(x):
    if x is not None:
        ind = str(x).find("#")
        if ind == -1:
            comment = str(x)
        else:
            comment = str(x).strip()
    else:
        comment = ""
    if (comment != "")&(not(comment.startswith("#"))):
        comment = "#"+comment
    return comment
